/****************************************************************************
 *   Copyright (C) 2012 by Claas Anders "CaScAdE" Rathje                    *
 *   admiralcascade@gmail.com                                               *
 *   Project-URL: http://www.mylifesucks.de/oss/c-scale/                    *
 *                                                                          *
 *   This program is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 2 of the License.         *
 *                                                                          *
 *   This program is distributed in the hope that it will be useful,        *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of         *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for more details.                           *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to the                          *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include <avr/wdt.h>
#include "uart.h"

/***
 * Reading fuses address 0 to 1.. 0xE1, 0xD9 .. OK!
 * 1MHz internal osc
 */

// array length?
#define COUNT_OF(x) ((sizeof(x)/sizeof(0[x])) / ((size_t)(!(sizeof(x) % sizeof(0[x])))))

#define SEG_A 	 1
#define SEG_B 	 2
#define SEG_C 	 4
#define SEG_D 	 8
#define SEG_E 	16
#define SEG_F 	32
#define SEG_G 	64

#define ICON_BATT		  1
#define ICON_MINUS		  2
#define ICON_TARE		  4
#define ICON_HOLD		  8
#define ICON_KG			 16
#define ICON_LB			 32
#define ICON_OZ			 64
#define ICON_G			128


uint16_t samples[100];
uint8_t sampleCnt = 0;
uint8_t evaluate = 0;
uint8_t digitstate[] = { 0, 0, 0, 0, 0 };
uint8_t convertedDigit[] = { 0, 0, 0, 0, 0 };
char send_value_buffer[20];

uint8_t iconstate = 0;
uint8_t dotState = 0;
volatile uint8_t measure = 0;


volatile uint8_t no_standby = 0;
uint8_t pre_no_standby_iconstate = 0;
uint8_t pre_no_standby_iconstate_checkme = 0;
uint16_t no_standby_prescaler = 0;
uint8_t send_value_precaler = 0;
volatile uint8_t send_value = 0;

// each row is pulsed for ~ 1.5ms
// => 500us is a good sample rate
void init_timer1(void) {
	TCCR1B |= (1 << WGM12); // CTC Modus
	TCCR1B |= (1 << CS10); // Prescaler 0
	OCR1A = 500;

	// Enable comapre Interrupt
	TIMSK |= (1 << OCIE1A);
}

/**
 * timer2 compare handler
 * -> 500us
 */
SIGNAL(TIMER1_COMPA_vect) {
	measure = 1;

	// 5 sec
	if (no_standby_prescaler++ == (30 * 2000UL)) {
		no_standby = 1;
		no_standby_prescaler = 0;
	}

	// 100ms
	if (send_value_precaler++ == 200) {
		send_value = 1;
		send_value_precaler = 0;
	}
}

/**
 * reverse decode 7segment character
 */
uint8_t decodeSegment(uint8_t segments) {
	switch (segments) {
	case 0b00111111:
		return 0x00;
	case 0b00000110:
		return 0x01;
	case 0b01011011:
		return 0x02;
	case 0b01001111:
		return 0x03;
	case 0b01100110:
		return 0x04;
	case 0b01101101:
		return 0x05;
	case 0b01111101:
		return 0x06;
	case 0b00000111:
		return 0x07;
	case 0b01111111:
		return 0x08;
	case 0b01101111:
		return 0x09;
	case 0b01110111:
		return 0x0A;
	case 0b01111100:
		return 0x0B;
	case 0b00111001:
		return 0x0C;
	case 0b01011110:
		return 0x0D;
	case 0b01111001:
		return 0x0E;
	case 0b01110001:
		return 0x0F;
	}
	return 0xFF; // fail
}

void pressMode() {
	DDRC |= (1 << PC2); // output
	PORTC &= ~(1 << PC2); // connect to gnd

	received = 0;
	_delay_ms(30);

	DDRC &= ~(1 << PC2); // input
	PORTC &= ~(1 << PC2); // NO pullup -> triste
}

void pressOnOff() {
	DDRC |= (1 << PC3); // output
	PORTC &= ~(1 << PC3); // connect to gnd

	received = 0;
	_delay_ms(30);

	DDRC &= ~(1 << PC3); // input
	PORTC &= ~(1 << PC3); // NO pullup -> triste
}

int main() {

	// mode
	DDRC &= ~(1 << PC2); // input
	PORTC &= ~(1 << PC2); // NO pullup -> triste

	// on/off
	DDRC &= ~(1 << PC3); // input
	PORTC &= ~(1 << PC3); // NO pullup -> triste

	// 16 inputs
	DDRB &= ~((1 << PB0) | (1 << PB1) | (1 << PB2) | (1 << PB3) | (1 << PB4)
			| (1 << PB5) | (1 << PB6) | (1 << PB7));
	// PD0 and PD1 are for UART
	DDRD &= ~((1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5) | (1 << PD6)
			| (1 << PD7));
	DDRC &= ~((1 << PC0) | (1 << PC1));

	init_timer1();
	uart_init();

	uart_puts("\x1B[2J\x1B[H");
	uart_puts_pgm(PSTR("C-Scale\r\n"));

	sei();

	while (1) {

		// prevent the scale from going to standby
		if (no_standby && !pre_no_standby_iconstate_checkme) {
			uint8_t units = iconstate & (ICON_KG | ICON_G | ICON_OZ	| ICON_LB);
			if ((units & ~ICON_KG) == 0 || (units & ~ICON_G) == 0 || (units & ~ICON_OZ) == 0 || (units & ~ICON_LB) == 0) {
				pre_no_standby_iconstate = iconstate;
				if (pre_no_standby_iconstate & (ICON_KG | ICON_G | ICON_OZ
						| ICON_LB)) {
					pre_no_standby_iconstate_checkme = 1;
					pressMode();
					_delay_ms(30);
					pressMode();
					_delay_ms(30);
					pressMode();
				}
			}
			no_standby = 0;
		}

		// handle received characters
		if (received == 'm') {
			pressMode();
		} else if (received == 'o') {
			pressOnOff();
		}

		// in case we do not need to evaluate we can sample the display pins
		if (measure && !evaluate) {
			measure = 0;
			uint16_t bits = PINB;
			bits |= ((PIND & ((1 << PD2) | (1 << PD3) | (1 << PD4) | (1 << PD5)
					| (1 << PD6) | (1 << PD7))) << 8);
			bits |= ((PINC & ((1 << PC0) | (1 << PC1))) << 8);

			// single row is valid capture
			if ((bits & 0b00001111) == 1 || (bits & 0b00001111) == 2 || (bits
					& 0b00001111) == 4 || (bits & 0b00001111) == 8) {
				if (sampleCnt < COUNT_OF(samples)) {
					samples[sampleCnt++] = bits;
				} else {
					evaluate = 1;
				}
			} else if (sampleCnt > 10) {
				evaluate = 1;
			} else {
				// invalid capture... maybe reset?
				sampleCnt = 0;
			}
		} else if (evaluate) { // captured enough samples
			for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
				digitstate[i] = 0;
				convertedDigit[i] = 0xFF;
			}
			iconstate = 0;
			dotState = 0;

			for (uint8_t s = 0; s < sampleCnt; s++) {
				uint16_t bits = samples[s];
				// row 
				if (bits & (1 << 0)) {
					// possible digits
					for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
						if (!(bits & (1 << (15 - i * 2)))) {
							digitstate[i] |= SEG_A;
						}
						if (!(bits & (1 << (14 - i * 2)))) {
							digitstate[i] |= SEG_F;
						}
					}

					// could be battery				
					if (!(bits & (1 << 4))) {
						iconstate |= ICON_BATT;
					}
					// could be g
					if (!(bits & (1 << 5))) {
						iconstate |= ICON_G;
					}
				}
				// row
				if (bits & (1 << 1)) {
					// possible digits
					for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
						if (!(bits & (1 << (15 - i * 2)))) {
							digitstate[i] |= SEG_B;
						}
						if (!(bits & (1 << (14 - i * 2)))) {
							digitstate[i] |= SEG_G;
						}
					}
					// could be minus				
					if (!(bits & (1 << 4))) {
						iconstate |= ICON_MINUS;
					}
					// could be lb
					if (!(bits & (1 << 5))) {
						iconstate |= ICON_LB;
					}
				}
				// row
				if (bits & (1 << 2)) {
					// possible digits
					for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
						if (!(bits & (1 << (15 - i * 2)))) {
							digitstate[i] |= SEG_C;
						}
						if (!(bits & (1 << (14 - i * 2)))) {
							digitstate[i] |= SEG_E;
						}
					}
					// could be TARE				
					if (!(bits & (1 << 4))) {
						iconstate |= ICON_TARE;
					}
					// could be oz
					if (!(bits & (1 << 5))) {
						iconstate |= ICON_OZ;
					}
				}
				// row
				if (bits & (1 << 3)) {
					// possible digits
					for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
						if (!(bits & (1 << (14 - i * 2)))) {
							digitstate[i] |= SEG_D;
						}
					}
					// could be dot
					for (uint8_t i = 1; i < COUNT_OF(digitstate) - 1; i++) {
						if (!(bits & (1 << (15 - i * 2)))) {
							dotState |= (1 << i);
						}
					}

					// could be HOLD				
					if (!(bits & (1 << 7))) {
						iconstate |= ICON_HOLD;
					}
					// could be kg
					if (!(bits & (1 << 5))) {
						iconstate |= ICON_LB;
					}
				}

			}


			// check if only valid numbers have been captured
			// this fails on "0-Ld" but that could be added if wanted
			uint8_t valid = 1;
			for (uint8_t i = 0; i < COUNT_OF(digitstate); i++) {
				convertedDigit[i] = decodeSegment(digitstate[i]);
				if (convertedDigit[i] == 0xFF && digitstate[i] != 0) {
					valid = 0;
				} else if (digitstate[i] != 0) {
					//value += (uint16_t)convertedDigit[i] * factor;
				}
				//value = digitstate[1];
				//factor *= 10;
			}


			uint8_t	units = iconstate & (ICON_KG | ICON_G | ICON_OZ	| ICON_LB);

			// no unit shown? -> invalid capture
			valid = valid && units;

			// more than one unit shown? -> invalid capture
			//valid = valid && ((units & ~ICON_KG) == 0 || (units & ~ICON_G) == 0 || (units & ~ICON_OZ) == 0 || (units & ~ICON_LB) == 0);
			
			
			if (valid) {
				// check if scale still shows same unit as prior to a standy-prevention 
				if (pre_no_standby_iconstate_checkme) {
					if ((pre_no_standby_iconstate & (ICON_KG | ICON_G | ICON_OZ
							| ICON_LB)) == (iconstate & (ICON_KG | ICON_G
							| ICON_OZ | ICON_LB))) {
						pre_no_standby_iconstate_checkme = 0;
					} else {
						/*char itoabuf[12];
						 itoa(pre_no_standby_iconstate_checkme, itoabuf, 10 );
						 uart_puts(itoabuf);	
						 uart_puts(" -> ");
						 itoa(iconstate, itoabuf, 10 );
						 uart_puts(itoabuf);	
						 uart_puts("\r\n");*/
						pressMode();
						_delay_ms(100);

						// and check again
					}

				}

				// unit is correct
				if ((!pre_no_standby_iconstate_checkme)) {

					// only update buffer if two measures in a row
					// yield in the same state
					// the scale tend to flicker from time to time
					// with a random delta for single frames
					// -> not visible on display but causes big jumps
					// in recorded values
					static uint8_t state_checksum = 255;
					uint8_t checksum = 0;
					for (int8_t i = 0; i < COUNT_OF(convertedDigit); i++) {
						checksum += convertedDigit[i];
					}

					checksum += iconstate;
					checksum += dotState;
					checksum = (checksum % 254); // 255 is init flag

					if (checksum == state_checksum || checksum == 255) {
						uint8_t bufferIndex = 0;
						if (iconstate & ICON_MINUS) {
							send_value_buffer[bufferIndex++] = '-';
						} else {
							send_value_buffer[bufferIndex++] = ' ';
						}

						for (int8_t i = COUNT_OF(convertedDigit) - 1; i >= 0; i--) {
							if (convertedDigit[i] == 0xFF) {
								send_value_buffer[bufferIndex++] = ' ';
							} else {
								send_value_buffer[bufferIndex++] = 0x30
										+ convertedDigit[i];
							}
							if (dotState & (1 << i)) {
								send_value_buffer[bufferIndex++] = '.';
							}
						}

						/*send_value_buffer[bufferIndex++] = '>';
						 itoa(dotState, send_value_buffer + bufferIndex, 10 );
						 bufferIndex += strlen(send_value_buffer + bufferIndex);
						 send_value_buffer[bufferIndex++] = '<';*/
						if (iconstate & ICON_G) {
							send_value_buffer[bufferIndex++] = 'g';
						} else if (iconstate & ICON_OZ) {
							send_value_buffer[bufferIndex++] = 'o';
							send_value_buffer[bufferIndex++] = 'z';
						} else if (iconstate & ICON_LB) {
							send_value_buffer[bufferIndex++] = 'l';
							send_value_buffer[bufferIndex++] = 'b';
						} else if (iconstate & ICON_KG) {
							send_value_buffer[bufferIndex++] = 'k';
							send_value_buffer[bufferIndex++] = 'g';
						}

						send_value_buffer[bufferIndex++] = '\r';
						send_value_buffer[bufferIndex++] = '\n';
						send_value_buffer[bufferIndex++] = '\0';
					}
					state_checksum = checksum;
				}

			} else {
				memset(send_value_buffer, 0, COUNT_OF(send_value_buffer));
			}

			evaluate = 0;
			sampleCnt = 0;
		}
		// time to send out new value
		if (send_value) {
			int8_t len = strlen((char*) send_value_buffer) - 1;
			if (len > 0) {
				while (TxData0Len > 0)
					; // wait for previous	
				TxData0Len = TXD0_BLOCKED;
				memcpy((void *) txd0_buffer, send_value_buffer,
						COUNT_OF(send_value_buffer));
				pTxData0 = txd0_buffer;
				TxData0Len = len;
				uart_putc(*pTxData0++);
				
				// only on valid send... 
				send_value = 0;
			}

		}

	} // while (1)

}
